from ...Structure import Limit

class LimitBuilder :
    """Builder component for LIMIT and OFFSET query
    """

    def __init__(self) :
        self.__limit = None

    def getLimit(self) -> tuple :
        """Get limit and offset number"""
        return self.__limit

    def hasLimit(self) -> bool :
        """Check limit or offset defined"""
        if self.__limit is None : return False
        else : return True

    def setLimit(self, limit: Limit) :
        """Set limit and offset number"""
        self.__limit = limit



from ..Structure import Table, Column, Value, Expression

class BaseBuilder :
    """Base of builder object.
    Basic template for building a query
    """

    SELECT = 1
    INSERT = 2
    UPDATE = 3
    DELETE = 4
    SELECT_DISTINCT = 5
    SELECT_UNION = 6
    SELECT_INTERSECT = 7
    SELECT_MINUS = 8
    INSERT_COPY = 9

    def __init__(self) :
        self.__builderType = 0
        self.__table = None
        self.__columns = ()
        self.__values = ()

    def builderType(self, type: int = 0) -> int :
        """Get or set builder object type"""
        if (type > 0 and type <= 9) :
            self.__builderType = type
        return self.__builderType

    def getTable(self) -> Table :
        """Get table of a builder"""
        return self.__table

    def setTable(self, table: Table) :
        """Set table for a builder"""
        self.__table = table

    def getColumns(self) -> tuple :
        """Get list of columns"""
        return self.__columns

    def countColumns(self) -> int :
        """Count column list"""
        return len(self.__columns)

    def addColumn(self, column) :
        """Add a column or column expression to Column list"""
        if isinstance(column, Column) or isinstance(column, Expression) :
            self.__columns = self.__columns + (column,)

    def getValues(self) -> tuple :
        """Get Value list"""
        return self.__values

    def countValues(self) -> int :
        """Count Value list"""
        return len(self.__values)

    def addValue(self, value : Value) :
        """Add a value to Value list"""
        self.__values = self.__values + (value,)



from ...Structure import Column

class GroupByBuilder :
    """Builder component for GROUP BY query
    """

    def __init__(self) :
        self.__groupBy = ()

    def getGroup(self) -> tuple :
        """Get list of column group"""
        return self.__groupBy

    def countGroup(self) -> int :
        """Count list of column group"""
        return len(self.__groupBy)

    def addGroup(self, groupBy: Column) :
        """Add a column to group list"""
        self.__groupBy = self.__groupBy + (groupBy,)



from .Column import Column
from .Expression import Expression
from typing import Iterable

class Clause :
    """Object for storing a clause query structure. Used in WHERE and HAVING query.
    Object properties:
    - Column or Expression object
    - Clause operator
    - Clause comparison value
    - Clause conjunctive
    - Nested level
    """

    NONE = 0
    WHERE = 1
    HAVING = 2

    OPERATOR_DEFAULT = 0
    OPERATOR_EQUAL = 1
    OPERATOR_NOT_EQUAL = 2
    OPERATOR_GREATER = 3
    OPERATOR_GREATER_EQUAL = 4
    OPERATOR_LESS = 5
    OPERATOR_LESS_EQUAL = 6
    OPERATOR_LIKE = 7
    OPERATOR_NOT_LIKE = 8
    OPERATOR_BETWEEN = 9
    OPERATOR_NOT_BETWEEN = 10
    OPERATOR_IN = 11
    OPERATOR_NOT_IN = 12
    OPERATOR_NULL = 13
    OPERATOR_NOT_NULL = 14

    CONJUNCTIVE_NONE = 0
    CONJUNCTIVE_AND = 1
    CONJUNCTIVE_OR = 2
    CONJUNCTIVE_NOT_AND = 3
    CONJUNCTIVE_NOT_OR = 4

    clauseType = NONE
    nestedConjunctive = CONJUNCTIVE_NONE
    nestedLevel = 0

    def __init__(self, column, operator: int, value, conjunctive: int, level: int) :
        self.__column = column
        if operator > 0 and operator <= 14 :
            self.__operator = operator
        else :
            self.__operator = self.OPERATOR_DEFAULT
        self.__value = value
        if conjunctive > 0 and conjunctive <= 4 :
            self.__conjunctive = conjunctive
        else :
            self.__conjunctive = self.CONJUNCTIVE_NONE
        self.__level = level

    def column(self) :
        """Get clause column. Can be column object or expression object."""
        return self.__column

    def operator(self) -> int :
        """Get operator type."""
        return self.__operator

    def value(self) :
        """Get clause comparison value."""
        return self.__value

    def conjunctive(self) -> int :
        """Get clause conjunctive type."""
        return self.__conjunctive

    def level(self, input: int = -1) -> int :
        """Get or set nested clause level. Negative for open parenthesis and positive for close parenthesis."""
        if input > -1 : self.__level = input
        return self.__level

    @classmethod
    def create(cls, clauseType: int, column, operator, value, conjunctive: int) :
        """Create Clause object from column input, operator, value, and conjunctive for WHERE or HAVING query."""
        if isinstance(column, Expression) or isinstance(column, Column) :
            columnObject = column
        else :
            columnObject = Column.create(column)
        validOperator = cls.getOperator(operator)
        validValue = cls.getValue(value, validOperator)
        conjunctive = cls.getConjunctive(clauseType, conjunctive)
        nestedLevel = cls.nestedLevel
        cls.clauseType = clauseType
        cls.nestedLevel = 0
        return Clause(columnObject, validOperator, validValue, conjunctive, nestedLevel)

    @classmethod
    def getOperator(cls, operator) -> int :
        """Get a valid operator option from input operator."""
        if isinstance(operator, int) :
            validOperator = operator
        else :
            if operator == '=' or operator == '==' :
                validOperator = Clause.OPERATOR_EQUAL
            elif operator == '!=' or operator == '<>' :
                validOperator = Clause.OPERATOR_NOT_EQUAL
            elif operator == '>' :
                validOperator = Clause.OPERATOR_GREATER
            elif operator == '>=' :
                validOperator = Clause.OPERATOR_GREATER_EQUAL
            elif operator == '<' :
                validOperator = Clause.OPERATOR_LESS
            elif operator == '<=' :
                validOperator = Clause.OPERATOR_LESS_EQUAL
            elif operator == 'BETWEEN' :
                validOperator = Clause.OPERATOR_BETWEEN
            elif operator == 'NOT BETWEEN' :
                validOperator = Clause.OPERATOR_NOT_BETWEEN
            elif operator == 'LIKE' :
                validOperator = Clause.OPERATOR_LIKE
            elif operator == 'NOT LIKE' :
                validOperator = Clause.OPERATOR_NOT_LIKE
            elif operator == 'IN' :
                validOperator = Clause.OPERATOR_IN
            elif operator == 'NOT IN' :
                validOperator = Clause.OPERATOR_NOT_IN
            elif operator == 'NULL' or operator == 'IS NULL' :
                validOperator = Clause.OPERATOR_NULL
            elif operator == 'NOT NULL' or operator == 'IS NOT NULL' :
                validOperator = Clause.OPERATOR_NOT_NULL
            else :
                validOperator = Clause.OPERATOR_DEFAULT
        return validOperator

    @classmethod
    def getValue(cls, value, operator: int) :
        """Check and get valid clause comparison value."""
        valid = True
        if operator == Clause.OPERATOR_BETWEEN or operator == Clause.OPERATOR_NOT_BETWEEN :
            if isinstance(value, Iterable) :
                valid = len(value) == 2
        if operator == Clause.OPERATOR_IN or operator == Clause.OPERATOR_NOT_IN :
            valid = isinstance(value, Iterable)
        if valid :
            return value
        else :
            raise Exception('Invalid input value for Where or Having clause')

    @classmethod
    def getConjunctive(cls, clauseType: int, conjunctive: int) -> int :
        """Get appropriate conjunctive from input conjunctive."""
        if clauseType == cls.clauseType :
            if conjunctive == Clause.CONJUNCTIVE_NONE :
                if cls.nestedConjunctive == Clause.CONJUNCTIVE_NONE : return Clause.CONJUNCTIVE_AND
                else : return cls.nestedConjunctive
            else :
                return conjunctive
        else :
            return Clause.CONJUNCTIVE_NONE



from .BaseQuery import BaseQuery
from .Component import Clauses, Where, LimitOffset
from ..Builder import BaseBuilder, DeleteBuilder
from ..Structure import Table

class Delete(BaseQuery, Clauses, Where, LimitOffset) :
    """DELETE query manipulation class.
    Components:
    - Table
    - Where clause
    - Limit query
    """

    def __init__(self, translator: int, bindingOption: int) :
        BaseQuery.__init__(self)
        self.builder = DeleteBuilder()
        self.builder.builderType(BaseBuilder.DELETE)
        self.translator = translator
        self.bindingOption = bindingOption
        Clauses.__init__(self)

    def delete(self, table) :
        """DELETE query table input"""
        if table :
            tableObject = Table.create(table)
            self.builder.setTable(tableObject)
        else :
            raise Exception("Table name is not defined")
        return self



from .Clauses import Clauses
from .Where import Where
from .Having import Having
from .GroupBy import GroupBy
from .OrderBy import OrderBy
from .LimitOffset import LimitOffset
from .JoinTable import JoinTable


"""Query translator class.
Interfacing query class with translator class and translate query object to query string and parameters.
"""

from .Constant import *
from .QueryObject import QueryObject
from .Builder import SelectBuilder, InsertBuilder, UpdateBuilder, DeleteBuilder
from .Translator import GenericTranslator, MySQLTranslator

def translateBuilder(query: QueryObject, builder, translator: int) :
    """Translate builder object to query object using selected translator"""
    translatorClass = getTranslator(query, translator)
    if isinstance(builder, SelectBuilder) :
        translatorClass.translateSelect(query, builder)
    elif isinstance(builder, InsertBuilder) :
        translatorClass.translateInsert(query, builder)
    elif isinstance(builder, UpdateBuilder) :
        translatorClass.translateUpdate(query, builder)
    elif isinstance(builder, DeleteBuilder) :
        translatorClass.translateDelete(query, builder)
    else :
        raise Exception('Tried to translate unregistered builder object')

def getTranslator(query: QueryObject, translator: int) :
    """Get selected translator instance"""
    if translator == TRANSLATOR_GENERIC :
        return GenericTranslator(query)
    elif translator == TRANSLATOR_MYSQL :
        return MySQLTranslator(query)
    else :
        raise Exception('Translator selected is not registered')

def getBindingOption(bindingOption: int) :
    """Get binding flag and mode from option"""
    bindingFlag = False
    bindingMode = False
    if bindingOption == PARAM_ASSOC :
        bindingFlag = True
        bindingMode = True
    elif bindingOption == PARAM_NUM :
        bindingFlag = True
        bindingMode = False
    return (bindingFlag, bindingMode)

def getQuery(query: QueryObject, bindingOption: int) -> str :
    """Get query string from input query object with binding option"""
    (bindingFlag, bindingMode) = getBindingOption(bindingOption)
    queryString = ''
    parts = query.parts()
    params = query.params()
    if bindingMode : mark = query.bindMarkAssoc()
    else : mark = query.bindMarkNum()
    quote = query.stringQuote()
    for i in range(len(parts)) :
        queryString += parts[i]
        if i < len(params) :
            if bindingFlag :
                if bindingMode : queryString += (mark + 'v' + str(i))
                else : queryString += mark
            else :
                if isinstance(params[i], str) :
                    queryString += (quote + params[i] + quote)
                elif isinstance(params[i], bytes) or isinstance(params[i], bytearray) :
                    queryString += (quote + str(params[i], 'utf-8') + quote)
                else :
                    queryString += str(params[i])
    return queryString

def getParams(query: QueryObject, bindingOption: int) :
    """Get params from input query object with binding option"""
    (bindingFlag, bindingMode) = getBindingOption(bindingOption)
    array = None
    if bindingFlag :
        if bindingMode :
            array = {}
            for i, param in enumerate(query.params()) :
                array['v' + str(i)] = param
        else :
            array = query.params()
    return array



from ...Structure import Clause, Expression
from ...Builder import HavingBuilder
from typing import Iterable

class Having :
    """HAVING clause manipulation component.
    Used for HAVING query.
    """

    def beginHaving(self) :
        """Begin a nested HAVING clause"""
        Clause.nestedConjunctive = Clause.CONJUNCTIVE_NONE
        Clause.nestedLevel -= 1
        return self

    def beginAndHaving(self) :
        """Begin a nested HAVING clause with AND conjunctive"""
        Clause.nestedConjunctive = Clause.CONJUNCTIVE_AND
        Clause.nestedLevel -= 1
        return self

    def beginOrHaving(self) :
        """Begin a nested HAVING clause with OR conjunctive"""
        Clause.nestedConjunctive = Clause.CONJUNCTIVE_OR
        Clause.nestedLevel -= 1
        return self

    def beginNotAndHaving(self) :
        """Begin a nested HAVING clause with NOT AND conjunctive"""
        Clause.nestedConjunctive = Clause.CONJUNCTIVE_NOT_AND
        Clause.nestedLevel -= 1
        return self

    def beginNotOrHaving(self) :
        """Begin a nested HAVING clause with NOT OR conjunctive"""
        Clause.nestedConjunctive = Clause.CONJUNCTIVE_NOT_OR
        Clause.nestedLevel -= 1
        return self

    def endHaving(self) :
        """End a nested HAVING clause"""
        if isinstance(self.builder, HavingBuilder) :
            lastClause = self.builder.lastHaving()
            if lastClause is not None :
                lastLevel = lastClause.level()
                lastClause.level(lastLevel + 1)
        return self

    def __having(self, column, operator, value, conjunctive: int) :
        """Add Clause object to having property of Builder object"""
        clauseObject = Clause.create(Clause.HAVING, column, operator, value, conjunctive)
        if isinstance(self.builder, HavingBuilder) :
            self.builder.addHaving(clauseObject)
        else :
            raise Exception('Builder object does not support HAVING query')
        return self

    def having(self, column, operator: str, value = None) :
        """HAVING query without conjunctive"""
        return self.__having(column, operator, value, Clause.CONJUNCTIVE_NONE)

    def andHaving(self, column, operator: str, value = None) :
        """HAVING query with AND conjunctive"""
        return self.__having(column, operator, value, Clause.CONJUNCTIVE_AND)

    def orHaving(self, column, operator: str, value = None) :
        """HAVING query with OR conjunctive"""
        return self.__having(column, operator, value, Clause.CONJUNCTIVE_OR)

    def notAndHaving(self, column, operator: str, value = None) :
        """HAVING query with NOT AND conjunctive"""
        return self.__having(column, operator, value, Clause.CONJUNCTIVE_NOT_AND)

    def notOrHaving(self, column, operator: str, value = None) :
        """HAVING query with NOT OR conjunctive"""
        return self.__having(column, operator, value, Clause.CONJUNCTIVE_NOT_OR)

    def havingExpression(self, expression, operator, value, params: Iterable = ()) :
        """HAVING expression query with AND conjunctive"""
        expressionObject = Expression.create(expression, '', params)
        return self.__having(expressionObject, operator, value, Clause.CONJUNCTIVE_AND)

    def orHavingExpression(self, expression, operator, value, params: Iterable = ()) :
        """HAVING expression query with OR conjunctive"""
        expressionObject = Expression.create(expression, '', params)
        return self.__having(expressionObject, operator, value, Clause.CONJUNCTIVE_OR)

    def notHavingExpression(self, expression, operator, value, params: Iterable = ()) :
        """HAVING expression query with NOT AND conjunctive"""
        expressionObject = Expression.create(expression, '', params)
        return self.__having(expressionObject, operator, value, Clause.CONJUNCTIVE_NOT_AND)

    def notOrHavingExpression(self, expression, operator, value, params: Iterable = ()) :
        """HAVING expression query with NOT OR conjunctive"""
        expressionObject = Expression.create(expression, '', params)
        return self.__having(expressionObject, operator, value, Clause.CONJUNCTIVE_NOT_OR)



from ...Structure import Column, Order
from ...Builder import OrderByBuilder

class OrderBy :
    """ORDER BY manipulation component.
    """

    def oderBy(self, columns, orderType) :
        """ORDER BY query manipulation"""
        for columnObject in Column.createMulti(columns) :
            orderObject = Order.create(columnObject, orderType)
            if isinstance(self.builder, OrderByBuilder) :
                self.builder.addOrder(orderObject)
            else :
                raise Exception('Builder object does not support ORDER BY query')
        return self

    def orderByAsc(self, columns) :
        """ORDER BY query with ASC order"""
        return self.oderBy(columns, Order.ORDER_ASC)

    def orderByDesc(self, columns) :
        """ORDER BY query with DESC order"""
        return self.oderBy(columns, Order.ORDER_DESC)



from typing import Iterable

class Expression :
    """Object for storing a user custom query expression.
    Object properties:
    - Query expression
    - Expression alias
    - Expression parameters
    """

    def __init__(self, expression: tuple, alias: str = '', params: tuple = ()) :
        countExp = len(expression)
        countPar = len(params)
        if countExp < countPar :
            params = params[:countExp]
        if countExp > countPar :
            for i in range(countPar, countExp) : params += (None,)
        self.__expression = expression
        self.__alias = alias
        self.__params = params

    def expression(self) -> tuple :
        """Get array of expression string."""
        return self.__expression

    def alias(self) -> str :
        """Get alias name of expression."""
        return self.__alias

    def params(self) -> tuple :
        """Get expression parameters array."""
        return self.__params

    @classmethod
    def create(cls, expression, alias = '', params: Iterable = ()) :
        """Create Expression object used in column list, where or having clause column, or group by column."""
        exps = ()
        if isinstance(expression, str) :
            exps = expression.split('?')
        elif isinstance(expression, Iterable) :
            exps = tuple(expression)
        return Expression(exps, str(alias), params)



class QueryObject :
    """An object for storing translated query.
    Contain list of query string parts and query parameters
    """

    def __init__(self) :
        """Clear query parts and all parameters"""
        self.__parts = ()
        self.__params = ()
        self.__number = -1
        self.__paramSet = True
        self.__bindMarkNum = '?'
        self.__bindMarkAssoc = ':'
        self.__stringQuote = '\''

    def parts(self) -> tuple :
        """Get query parts list"""
        return self.__parts

    def params(self) -> tuple :
        """Get value parameters list"""
        return self.__params

    def add(self, queryPart, paramFlag: bool = False) :
        """Add a query part or a query parameter to the list"""
        if not paramFlag or self.__number < 0 :
            if self.__paramSet :
                self.__parts += (queryPart,)
                self.__paramSet = False
                self.__number += 1
            else :
                self.__parts = self.__parts[:self.__number] + (self.__parts[self.__number] + queryPart,)
        else :
            self.__params += (queryPart,)
            self.__paramSet = True

    def bindMarkNum(self) -> str :
        """Get sequential binding mark"""
        return self.__bindMarkNum

    def bindMarkAssoc(self) -> str :
        """Get associative binding mark"""
        return self.__bindMarkAssoc

    def stringQuote(self) -> str :
        """Get string value quote character"""
        return self.__stringQuote

    def setMarkQuote(self, bindMarkNum: str, bindMarkAssoc: str, stringQuote: str) :
        """Set sequential and associative binding mark and quote character"""
        self.__bindMarkNum = bindMarkNum
        self.__bindMarkAssoc = bindMarkAssoc
        self.__stringQuote = stringQuote



from .BaseQuery import BaseQuery
from .Component import Clauses, Where, Having, GroupBy, OrderBy, LimitOffset, JoinTable
from ..Builder import BaseBuilder, SelectBuilder
from ..Structure import Table, Column, Expression
from typing import Iterable, Mapping

class Select(BaseQuery, Clauses, Where, Having, GroupBy, OrderBy, LimitOffset, JoinTable) :
    """SELECT query manipulation class.
    Components:
    - Table
    - Select columns
    - Join table query
    - Where clause
    - Group by query
    - Having clause
    - Order by query
    - Limit query
    """

    def __init__(self, translator: int, bindingOption: int) :
        BaseQuery.__init__(self)
        self.builder = SelectBuilder()
        self.builder.builderType(BaseBuilder.SELECT)
        self.translator = translator
        self.bindingOption = bindingOption
        Clauses.__init__(self)
        JoinTable.__init__(self)

    def select(self, table) :
        """SELECT query table input"""
        if table :
            tableObject = Table.create(table)
            self.builder.setTable(tableObject)
        else :
            raise Exception("Table name is not defined")
        return self

    def selectDistinct(self, table) :
        """SELECT DISTINC query table input"""
        self.builder.builderType(BaseBuilder.SELECT_DISTINCT)
        return self.select(table)

    def column(self, column) :
        """Add a column to builder object.
        
        Takes a column name string or a dictionary with keys as column alias.
        """
        columnObject = Column.create(column)
        self.builder.addColumn(columnObject)
        return self

    def columns(self, columns) :
        """Add multiple columns to builder object.
        
        Takes a list containing column name or a dictionary with keys as column alias.
        """
        for columnObject in Column.createMulti(columns) :
            self.builder.addColumn(columnObject)
        return self

    def columnExpression(self, expression, alias, params: Iterable = ()) :
        """Add an expression to list of Column in builder object.
        
        Takes expression string, expression alias, and list of parameters.
        """
        expressionObject = Expression.create(expression, alias, params)
        self.builder.addColumn(expressionObject)
        return self



from ...Structure import Clause

class HavingBuilder :
    """Builder component for HAVING clause query
    """

    def __init__(self) :
        self.__having = ()

    def getHaving(self) -> tuple :
        """Get list of having clause"""
        return self.__having

    def lastHaving(self) -> Clause :
        """Get last having clause in having list"""
        count = len(self.__having)
        if count > 0 :
            return self.__having[count-1]
        return None

    def countHaving(self) -> int :
        """Count list of having clause"""
        return len(self.__having)

    def addHaving(self, having: Clause) :
        """Add a having clause to having list"""
        self.__having += (having,)



from .WhereBuilder import WhereBuilder
from .HavingBuilder import HavingBuilder
from .GroupByBuilder import GroupByBuilder
from .OrderByBuilder import OrderByBuilder
from .LimitBuilder import LimitBuilder
from .JoinBuilder import JoinBuilder


from .. import QueryTranslator
from ..QueryObject import QueryObject

class BaseQuery :
    """Base of query manipulation class.
    Must be derived for a query manipulation class.
    Contain methods for translating query.
    """

    def __init__(self) :
        self.queryObject = None
        self.builder = None
        self.translator = 0
        self.bindingOption = 0
        self.queryFlag = False
        self.paramFlag = False

    def getQueryObject(self) :
        """Get current state of query object"""
        return self.queryObject

    def getBuilder(self) :
        """Get builder object"""
        return self.builder

    def translate(self, translator: int = 0) :
        """Translate builder object to query object"""
        if translator == 0 : translator = self.translator
        self.queryObject = QueryObject()
        QueryTranslator.translateBuilder(self.queryObject, self.builder, translator)

    def query(self, translator: int = 0, bindingOption: int = 0) :
        """Get query string of current state builder object"""
        if self.queryFlag == self.paramFlag :
            self.translate(translator)
        self.queryFlag = not self.queryFlag
        if bindingOption == 0 : bindingOption = self.bindingOption
        return QueryTranslator.getQuery(self.queryObject, bindingOption)

    def params(self, translator: int = 0, bindingOption: int = 0) :
        """Get query parameters of current state builder object"""
        if self.queryFlag == self.paramFlag :
            self.translate(translator)
        self.paramFlag = not self.paramFlag
        if bindingOption == 0 : bindingOption = self.bindingOption
        return QueryTranslator.getParams(self.queryObject, bindingOption)



from .Table import Table
from .Column import Column
from .Value import Value
from .Clause import Clause
from .Order import Order
from .Limit import Limit
from .Expression import Expression
from .Join import Join


from .Table import Table
from typing import Iterable, Mapping

class Value :
    """Object for storing a column and value pair of INSERT and UPDATE query.
    Object properties:
    - Table name or alias
    - List of column name
    - List of insert or update value
    """

    def __init__(self, table: str, columns: tuple, values: tuple) :
        self.__table = table
        lenCol = len(columns)
        lenVal = len(values)
        lenMin = lenCol
        if lenCol > lenVal : lenMin = lenVal
        self.__columns = columns[:lenMin]
        self.__values = values[:lenMin]

    def table(self) -> str :
        """Get table name of values"""
        return self.__table

    def columns(self) -> tuple :
        """Get column name of values"""
        return self.__columns

    def values(self) -> tuple :
        """Get values"""
        return self.__values

    @classmethod
    def create(cls, inputValue) :
        """Create Value object from ascossiative array with key as column or sequential array of array."""
        columns = ()
        values = ()
        if isinstance(inputValue, Mapping) :
            for key in inputValue.keys() :
                columns += (cls.getColumn(key),)
            values = tuple(inputValue.values())
        elif isinstance(inputValue, Iterable) :
            (columns, values) = cls.parsePair(inputValue)
        return Value(Table.table, columns, values)

    @classmethod
    def createMulti(cls, inputValues) :
        """Create multiple Value objects."""
        valuesObjects = ()
        if isinstance(inputValues, Iterable) :
            for val in inputValues :
                valuesObjects += (Value.create(val),)
        return valuesObjects

    @classmethod
    def parsePair(cls, pairs: Iterable) -> tuple :
        """Parsing column and value pair from input array."""
        if isinstance(pairs[0], str) and len(pairs) == 2 :
            return ((pairs[0],), (pairs[1],))
        columns = ()
        values = ()
        for pair in pairs :
            if isinstance(pair, Mapping) and len(pair) :
                key = next(iter(pair.keys()))
                columns += (cls.getColumn(key),)
                values += (pair[key],)
            elif isinstance(pair, Iterable) and len(pair) == 2 :
                columns += (cls.getColumn(pair[0]),)
                values += (pair[1],)
        return (columns, values)

    @classmethod
    def getColumn(cls, column) -> str :
        validColumn = ''
        if isinstance(column, str) :
            validColumn = column
        elif isinstance(column, bytes) or isinstance(column, bytearray) :
            validColumn = str(column, 'utf-8')
        return validColumn



from .BaseQuery import BaseQuery
from .Component import LimitOffset
from ..Builder import BaseBuilder, InsertBuilder
from ..Structure import Table, Value
from typing import Iterable

class Insert(BaseQuery, LimitOffset) :
    """INSERT query manipulation class.
    Components:
    - Table
    - Insert values
    - Limit query
    """

    def __init__(self, translator: int, bindingOption: int) :
        BaseQuery.__init__(self)
        self.builder = InsertBuilder()
        self.builder.builderType(BaseBuilder.INSERT)
        self.translator = translator
        self.bindingOption = bindingOption

    def insert(self, table) :
        """INSERT INTO query query table input"""
        if table :
            tableObject = Table.create(table)
            self.builder.setTable(tableObject)
        else :
            raise Exception("Table name is not defined")
        return self

    def values(self, values) :
        """Add value and column pair list to builder object.
        
        Takes a dictionary with keys as column or list of two list with first item as column.
        """
        valueObject = Value.create(values)
        self.builder.addValue(valueObject)
        return self

    def multiValues(self, multiValues) :
        """Add multiple value and column pair list to builder object.
        
        Takes list of dictionary with keys as column.
        """
        for valueObject in Value.createMulti(multiValues) :
            self.builder.addValue(valueObject)
        return self



class Limit :
    """Object for storing limit and offset of query operation. Used in LIMIT and OFFSET query
    Object properties:
    - Limit number
    - Offset number
    """

    NOT_SET = -1

    def __init__(self, limit: int, offset: int) :
        self.__limit = limit
        self.__offset = offset

    def limit(self) -> int :
        """Get limit number."""
        return self.__limit

    def offset(self) -> int :
        """Get offset number."""
        return self.__offset

    @classmethod
    def create(cls, limit, offset) :
        """Create Limit object from input limit and offset for LIMIT query."""
        validLimit = Limit.NOT_SET
        validOffset = Limit.NOT_SET
        if isinstance(limit, int) :
            if limit > 0 : validLimit = limit
        if isinstance(offset, int) :
            if offset > 0 : validOffset = offset
        return Limit(validLimit, validOffset)



from ...Structure import Join

class JoinBuilder :
    """Builder component for JOIN query
    """

    def __init__(self) :
        self.__join = ()

    def getJoin(self) -> tuple :
        """Get list of join table"""
        return self.__join

    def lastJoin(self) -> Join :
        """Get last join in where list"""
        count = len(self.__join)
        if count > 0 :
            return self.__join[count-1]
        return None

    def countJoin(self) -> int :
        """Count list of join table"""
        return len(self.__join)

    def addJoin(self, join: Join) :
        """Add a join to join list"""
        self.__join += (join,)



from ...Structure import Order

class OrderByBuilder :
    """Builder component for ORDER BY query
    """

    def __init__(self) :
        self.__orderBy = ()

    def getOrder(self) -> tuple :
        """Get query result order list"""
        return self.__orderBy

    def countOrder(self) -> int :
        """Count query result order list"""
        return len(self.__orderBy)

    def addOrder(self, orderBy: Order) :
        """Add a query result order to order list"""
        self.__orderBy = self.__orderBy + (orderBy,)



from .Table import Table
from typing import Iterable, Mapping

class Column :
    """Object for storing a column definition.
    Object properties:
    - Table name or alias
    - Column name
    - Column alias
    - Aggregate function
    """

    def __init__(self, table: str, name: str, function: str = '', alias: str = '') :
        self.__table = table
        self.__name = name
        self.__function = function
        self.__alias = alias

    def table(self) -> str :
        """Get table name or table alias of column."""
        return self.__table

    def name(self) -> str :
        """Get name of column."""
        return self.__name

    def function(self) -> str :
        """Get SQL aggregate function."""
        return self.__function

    def alias(self) -> str :
        """Get alias name of column."""
        return self.__alias

    @classmethod
    def create(cls, column) :
        """Create column object from string input or associative array with key as alias."""
        table = ''
        name = ''
        function = ''
        alias = ''
        if isinstance(column, str) or isinstance(column, bytes) or isinstance(column, bytearray) :
            (table, name, function) = cls.parseStr(column)
        elif isinstance(column, Mapping) :
            (table, name, function, alias) = cls.parseMap(column)
        return Column(table, name, function, alias)

    @classmethod
    def createMulti(cls, columns) :
        """Create multiple Column objects."""
        columnObjects = ()
        if isinstance(columns, str) or isinstance(columns, bytes) or isinstance(columns, bytearray) :
            columnObjects = (cls.create(columns),)
        elif isinstance(columns, Mapping) :
            for key in columns.keys() :
                columnObjects += (Column.create({key: columns[key]}),)
        elif isinstance(columns, Iterable) :
            for col in columns :
                columnObjects += (cls.create(col),)
        return columnObjects

    @classmethod
    def parseStr(cls, column) -> tuple :
        """Parsing string input column to table, column name, and aggregate function."""
        if isinstance(column, bytes) or isinstance(column, bytearray) :
            column = str(column, 'utf-8')
        elif not isinstance(column, str) :
            return ('', '', '')
        function = ''
        pos1 = column.find('(')
        pos2 = column.rfind(')')
        if pos1 > 0 and pos2 == len(column) - 1 :
            function = column[0:pos1]
            column = column[pos1+1:pos2]
        table = Table.table
        name = cls.dequote(column)
        split = column.split('.')
        length = len(split)
        if length == 2 :
            table = cls.dequote(split[0])
            name = cls.dequote(split[1])
        elif length > 2 :
            pos = column.find('"."')
            if pos <= 0 : pos = column.find("'.'")
            if pos <= 0 : pos = column.find('`.`')
            if pos > 0 :
                table = column[1:pos]
                name = column[pos+3:-1]
        return (table, name, function)

    @classmethod
    def parseMap(cls, column: Mapping) -> tuple :
        """Parsing array input column to table, column name, aggregate function, and alias name."""
        if len(column) == 1 :
            key = next(iter(column.keys()))
            alias = str(key)
            (table, name, function) = cls.parseStr(column[key])
            return (table, name, function, alias)
        else :
            return ('', '', '', '')

    @classmethod
    def dequote(cls, input: str) -> str :
        if (input[0] == input[-1]) and input.startswith(('"', "'", '`')) :
            return input[1:-1]
        return input



from ...Structure import Join
from ...Builder import JoinBuilder
from typing import Iterable

class JoinTable :
    """Join table manipulation component.
    Used for INNER JOIN, LEFT JOIN, RIGHT JOIN, OUTER JOIN query.
    """

    def __init__(self) :
        Join.table = ''

    def join(self, joinTable, joinType) :
        """Add Join object to join property of Builder object"""
        joinObject = Join.create(joinTable, joinType)
        if isinstance(self.builder, JoinBuilder) :
            self.builder.addJoin(joinObject)
        else :
            raise Exception('Builder object does not support JOIN query')
        return self

    def innerJoin(self, joinTable) :
        """INNER JOIN query. Takes join table name input"""
        return self.join(joinTable, Join.INNER_JOIN)

    def leftJoin(self, joinTable) :
        """LEFT JOIN query. Takes join table name input"""
        return self.join(joinTable, Join.LEFT_JOIN)

    def rightJoin(self, joinTable) :
        """RIGHT JOIN query. Takes join table name input"""
        return self.join(joinTable, Join.RIGHT_JOIN)

    def outerJoin(self, joinTable) :
        """OUTER JOIN query. Takes join table name input"""
        return self.join(joinTable, Join.OUTER_JOIN)

    def on(self, baseColumn, joinColumn) :
        """Add join columns and build ON query"""
        if isinstance(self.builder, JoinBuilder) :
            lastJoin = self.builder.lastJoin()
            if lastJoin is not None :
                lastJoin.addColumn(baseColumn, joinColumn)
        return self

    def using(self, columns) :
        """Add join columns and build USING query"""
        if isinstance(columns, str) or not isinstance(columns, Iterable) :
            columns = (columns,)
        if isinstance(self.builder, JoinBuilder) :
            lastJoin = self.builder.lastJoin()
            if lastJoin is not None :
                for column in columns :
                    lastJoin.addColumn(column)
        return self



"""SQL builder library.

Build a query using series of methods. 
Support build query for various RDBMS. 
Can produces a complete query string with input parameters or with separate parameters.

translator option:
- TRANSLATOR_GENERIC
- TRANSLATOR_BEAUTIFY
- TRANSLATOR_MYSQL
- TRANSLATOR_SQLITE

binding option:
- PARAM_NONE
- PARAM_NUM
- PARAM_ASSOC
"""

__version__ = '1.1.0'

from .Constant import *
from .Query import Select, Insert, Update, Delete

translator = TRANSLATOR_GENERIC
binding = PARAM_NUM

table = ''

def select(table_ = None) -> Select :
    """Begin SELECT query builder. Takes a database table as parameter"""
    if table_ is None : table_ = table
    selectQuery = Select(translator, binding)
    return selectQuery.select(table_)

def selectDistinct(table_ = None) -> Select :
    """Begin SELECT DISTINCT query builder. Takes a database table as parameter"""
    if table_ is None : table_ = table
    selectQuery = Select(translator, binding)
    return selectQuery.selectDistinct(table_)

def insert(table_ = None) -> Insert :
    """Begin INSERT query builder. Takes a database table as parameter"""
    if table_ is None : table_ = table
    insertQuery = Insert(translator, binding)
    return insertQuery.insert(table_)

def update(table_ = None) -> Update :
    """Begin UPDATE query builder. Takes a database table as parameter"""
    if table_ is None : table_ = table
    updateQuery = Update(translator, binding)
    return updateQuery.update(table_)

def delete(table_ = None) -> Delete :
    """Begin DELETE query builder. Takes a database table as parameter"""
    if table_ is None : table_ = table
    deleteQuery = Delete(translator, binding)
    return deleteQuery.delete(table_)



from ...Structure import Column
from ...Builder import GroupByBuilder

class GroupBy :
    """GROUP BY manipulation component.
    """

    def groupBy(self, columns) :
        """GROUP BY query manipulation"""
        for column in Column.createMulti(columns) :
            if isinstance(self.builder, GroupByBuilder) :
                self.builder.addGroup(column)
            else :
                raise Exception('Builder object does not support GROUP BY query')
        return self



from .Column import Column

class Order :
    """Object for storing order of query result definition. Used in ORDER BY query
    Object properties:
    - Column object
    - Oder type
    """

    ORDER_NONE = 0
    ORDER_ASC = 1
    ORDER_DESC = 2

    def __init__(self, column: Column, orderType: int) :
        self.__column = column
        self.__orderType = orderType

    def column(self) -> Column :
        """Get order column."""
        return self.__column

    def orderType(self) -> int :
        """Get order type."""
        return self.__orderType

    @classmethod
    def create(cls, column, orderType) :
        """Create Order object from column inputs and order type for ORDER BY query."""
        if isinstance(column, Column) :
            columnObject = column
        else :
            columnObject = Column.create(column)
        validType = cls.getOrderType(orderType)
        return Order(columnObject, validType)

    @classmethod
    def getOrderType(cls, orderType) -> int :
        """Get valid order type from input order type."""
        if isinstance(orderType, int) :
            validType = orderType
        else :
            if orderType == 'ASCENDING' or orderType == 'ASC' or orderType == 'ascending' or orderType == 'asc' :
                validType = Order.ORDER_ASC
            elif orderType == 'DESCENDING' or orderType == 'DESC' or orderType == 'descending' or orderType == 'desc' :
                validType = Order.ORDER_DESC
            else :
                validType = Order.ORDER_NONE
        return validType



from .BaseTranslator import BaseTranslator
from ..QueryObject import QueryObject
from ..Builder import SelectBuilder, InsertBuilder, UpdateBuilder, DeleteBuilder

class MySQLTranslator(BaseTranslator) :
    """Translator for MySQL database.
    """

    def __init__(self, query: QueryObject) :
        query.setMarkQuote("?", ":", "\"")
        self.quote_struct = "`"
        self.quote_string = "'"
        self.equal = "="
        self.open_bracket = "("
        self.close_bracket = ")"
        self.dot = "."
        self.comma = ", "
        self.end_query = ";"

    def translateSelect(self, query: QueryObject, builder: SelectBuilder) :
        """Translate SELECT query from select builder."""
        multiTableFlag = bool(builder.countJoin())
        self.firstKeyword(query, builder.builderType())
        self.columnsSelect(query, builder.getColumns(), builder.countColumns(), multiTableFlag)
        self.fromTable(query, builder.getTable())
        self.join(query, builder.getJoin())
        self.where(query, builder.getWhere(), builder.countWhere(), multiTableFlag)
        self.groupBy(query, builder.getGroup(), builder.countGroup(), multiTableFlag)
        self.having(query, builder.getHaving(), builder.countHaving(), multiTableFlag)
        self.orderBy(query, builder.getOrder(), builder.countOrder(), multiTableFlag)
        self.limitOffset(query, builder.getLimit(), builder.hasLimit())

    def translateInsert(self, query: QueryObject, builder: InsertBuilder) :
        """Translate INSERT query from insert builder."""
        self.firstKeyword(query, builder.builderType())
        self.intoTable(query, builder.getTable())
        self.columnsInsert(query, builder.getValues(), builder.countValues())
        self.valuesInsert(query, builder.getValues(), builder.countValues())
        self.limitOffset(query, builder.getLimit(), builder.hasLimit())

    def translateUpdate(self, query: QueryObject, builder: UpdateBuilder) :
        """Translate UPDATE query from update builder."""
        multiTableFlag = bool(builder.countJoin())
        self.firstKeyword(query, builder.builderType())
        self.tableSet(query, builder.getTable())
        self.join(query, builder.getJoin())
        self.valuesUpdate(query, builder.getValues(), builder.countValues(), multiTableFlag)
        self.where(query, builder.getWhere(), builder.countWhere(), multiTableFlag)
        self.limitOffset(query, builder.getLimit(), builder.hasLimit())

    def translateDelete(self, query: QueryObject, builder: DeleteBuilder) :
        """Translate DELETE query from delete builder."""
        self.firstKeyword(query, builder.builderType())
        self.fromTable(query, builder.getTable())
        self.where(query, builder.getWhere(), builder.countWhere())
        self.limitOffset(query, builder.getLimit(), builder.hasLimit())



from typing import Mapping

class Table :
    """Object for storing a table definition.
    Object properties:
    - Table name
    - Table name alias
    """

    table = ''

    def __init__(self, name: str, alias: str = '') :
        self.__name = name
        self.__alias = alias

    def name(self) -> str :
        """Get table name."""
        return self.__name

    def alias(self) -> str :
        """Get table alias."""
        return self.__alias

    @classmethod
    def create(cls, table) :
        """Create table object from string input or ascossiative array with key as alias."""
        name = ''
        alias = ''
        if isinstance(table, str) :
            name = table
            cls.table = table
        elif isinstance(table, bytes) or isinstance(table, bytearray) :
            name = str(table, 'utf-8')
            cls.table = name
        elif isinstance(table, Mapping) and len(table) == 1 :
            key = next(iter(table.keys()))
            name = str(table[key])
            alias = str(key)
            cls.table = alias
        return Table(name, alias)



from ...Structure import Clause

class WhereBuilder :
    """Builder component for WHERE clause query
    """

    def __init__(self) :
        self.__where = ()

    def getWhere(self) -> tuple:
        """Get list of where clause"""
        return self.__where

    def lastWhere(self) -> Clause :
        """Get last where clause in where list"""
        count = len(self.__where)
        if count > 0 :
            return self.__where[count-1]
        return None

    def countWhere(self) -> int :
        """Count list of where clause"""
        return len(self.__where)

    def addWhere(self, where: Clause) :
        """Add a where clause to where list"""
        self.__where += (where,)



from .BaseBuilder import BaseBuilder
from .Component import WhereBuilder, HavingBuilder, GroupByBuilder, OrderByBuilder, LimitBuilder, JoinBuilder

class SelectBuilder(BaseBuilder, WhereBuilder, HavingBuilder, GroupByBuilder, OrderByBuilder, LimitBuilder, JoinBuilder) :
    """Template for building SELECT query
    Components:
    - Where builder
    - Having builder
    - GroupBy builder
    - OrderBy builder
    - Limit builder
    - Join builder
    """

    def __init__(self) :
        BaseBuilder.__init__(self)
        WhereBuilder.__init__(self)
        HavingBuilder.__init__(self)
        GroupByBuilder.__init__(self)
        OrderByBuilder.__init__(self)
        LimitBuilder.__init__(self)
        JoinBuilder.__init__(self)



from .BaseQuery import BaseQuery
from .Component import Clauses, Where, LimitOffset, JoinTable
from ..Builder import BaseBuilder, UpdateBuilder
from ..Structure import Table, Value

class Update(BaseQuery, Clauses, Where, LimitOffset, JoinTable) :
    """UPDATE query manipulation class.
    Components:
    - Table
    - Update values
    - Where clause
    - Join table query
    - Limit query
    """

    def __init__(self, translator: int, bindingOption: int) :
        BaseQuery.__init__(self)
        self.builder = UpdateBuilder()
        self.builder.builderType(BaseBuilder.UPDATE)
        self.translator = translator
        self.bindingOption = bindingOption
        Clauses.__init__(self)
        JoinTable.__init__(self)

    def update(self, table) :
        """UPDATE query table input"""
        if table :
            tableObject = Table.create(table)
            self.builder.setTable(tableObject)
        else :
            raise Exception("Table name is not defined")
        return self

    def set(self, values) :
        """Add value and column pair set to builder object.
        
        Takes a dictionary with keys as column or list of two list with first item as column.
        """
        valueObject = Value.create(values)
        self.builder.addValue(valueObject)
        return self



from ...Structure import Clause, Expression
from ...Builder import WhereBuilder
from typing import Iterable

class Where :
    """WHERE clause manipulation component.
    Used for WHERE query.
    """

    def beginWhere(self) :
        """Begin a nested WHERE clause"""
        Clause.nestedConjunctive = Clause.CONJUNCTIVE_NONE
        Clause.nestedLevel -= 1
        return self

    def beginAndWhere(self) :
        """Begin a nested WHERE clause with AND conjunctive"""
        Clause.nestedConjunctive = Clause.CONJUNCTIVE_AND
        Clause.nestedLevel -= 1
        return self

    def beginOrWhere(self) :
        """Begin a nested WHERE clause with OR conjunctive"""
        Clause.nestedConjunctive = Clause.CONJUNCTIVE_OR
        Clause.nestedLevel -= 1
        return self

    def beginNotAndWhere(self) :
        """Begin a nested WHERE clause with NOT AND conjunctive"""
        Clause.nestedConjunctive = Clause.CONJUNCTIVE_NOT_AND
        Clause.nestedLevel -= 1
        return self

    def beginNotOrWhere(self) :
        """Begin a nested WHERE clause with NOT OR conjunctive"""
        Clause.nestedConjunctive = Clause.CONJUNCTIVE_NOT_OR
        Clause.nestedLevel -= 1
        return self

    def endWhere(self) :
        """End a nested WHERE clause"""
        if isinstance(self.builder, WhereBuilder) :
            lastClause = self.builder.lastWhere()
            if lastClause is not None :
                lastLevel = lastClause.level()
                lastClause.level(lastLevel + 1)
        return self

    def __where(self, column, operator, value, conjunctive: int) :
        """Add Clause object to where property of Builder object"""
        clauseObject = Clause.create(Clause.WHERE, column, operator, value, conjunctive)
        if isinstance(self.builder, WhereBuilder) :
            self.builder.addWhere(clauseObject)
        else :
            raise Exception('Builder object does not support WHERE query')
        return self

    def where(self, column, operator, value = None) :
        """WHERE query without conjunctive"""
        return self.__where(column, operator, value, Clause.CONJUNCTIVE_NONE)

    def andWhere(self, column, operator, value = None) :
        """WHERE query with AND conjunctive"""
        return self.__where(column, operator, value, Clause.CONJUNCTIVE_AND)

    def orWhere(self, column, operator, value = None) :
        """WHERE query with OR conjunctive"""
        return self.__where(column, operator, value, Clause.CONJUNCTIVE_OR)

    def notAndWhere(self, column, operator, value = None) :
        """WHERE query with NOT AND conjunctive"""
        return self.__where(column, operator, value, Clause.CONJUNCTIVE_NOT_AND)

    def notOrWhere(self, column, operator, value = None) :
        """WHERE query with NOT OR conjunctive"""
        return self.__where(column, operator, value, Clause.CONJUNCTIVE_NOT_OR)

    def whereExpression(self, expression, operator, value, params: Iterable = ()) :
        """WHERE expression query with AND conjunctive"""
        expressionObject = Expression.create(expression, '', params)
        return self.__where(expressionObject, operator, value, Clause.CONJUNCTIVE_AND)

    def orWhereExpression(self, expression, operator, value, params: Iterable = ()) :
        """WHERE expression query with OR conjunctive"""
        expressionObject = Expression.create(expression, '', params)
        return self.__where(expressionObject, operator, value, Clause.CONJUNCTIVE_OR)

    def notWhereExpression(self, expression, operator, value, params: Iterable = ()) :
        """WHERE expression query with NOT AND conjunctive"""
        expressionObject = Expression.create(expression, '', params)
        return self.__where(expressionObject, operator, value, Clause.CONJUNCTIVE_NOT_AND)

    def notOrWhereExpression(self, expression, operator, value, params: Iterable = ()) :
        """WHERE expression query with NOT OR conjunctive"""
        expressionObject = Expression.create(expression, '', params)
        return self.__where(expressionObject, operator, value, Clause.CONJUNCTIVE_NOT_OR)



from .BaseBuilder import BaseBuilder
from .Component import LimitBuilder

class InsertBuilder(BaseBuilder, LimitBuilder) :
    """Template for building INSERT query
    Components:
    - Limit builder
    """

    def __init__(self) :
        BaseBuilder.__init__(self)
        LimitBuilder.__init__(self)



from .Table import Table
from .Column import Column
from typing import Mapping

class Join :
    """Object for storing join table query definition. Used in INNER JOIN, LEFT JOIN, RIGHT JOIN, or OUTER JOIN query
    Object properties:
    - Join type
    - Base table name
    - Join table name
    - Join table alias name
    - List of base column objects
    - List of join column objects
    - List of using column objects
    """

    NO_JOIN = 0
    INNER_JOIN = 1
    LEFT_JOIN = 2
    RIGHT_JOIN = 3
    OUTER_JOIN = 4

    table = ''

    def __init__(self, joinType: int, baseTable: str, joinTable: str, joinAlias: str = '') :
        self.__joinType = joinType
        self.__baseTable = baseTable
        self.__joinTable = joinTable
        self.__joinAlias = joinAlias
        self.__baseColumns = ()
        self.__joinColumns = ()
        self.__usingColumns = ()

    def joinType(self) -> int :
        """Get join type."""
        return self.__joinType

    def baseTable(self) -> str :
        """Get base table name."""
        return self.__baseTable

    def joinTable(self) -> str :
        """Get join table name."""
        return self.__joinTable

    def joinAlias(self) -> str :
        """Get join table alias name."""
        return self.__joinAlias

    def baseColumns(self) -> tuple :
        """Get base column objects list."""
        return self.__baseColumns

    def joinColumns(self) -> tuple :
        """Get join column objects list"""
        return self.__joinColumns

    def usingColumns(self) -> tuple :
        """Get using column objects list"""
        return self.__usingColumns

    @classmethod
    def create(cls, joinTable, joinType) :
        """Create join table object from joint type and input table"""
        cls.table = Table.table
        validType = cls.getType(joinType)
        if isinstance(joinTable, Mapping) :
            key = next(iter(joinTable.keys()))
            joinAlias = str(key)
            joinObject = Join(validType, Table.table, cls.getTable(joinTable[key]), joinAlias)
            Table.table = joinAlias
        else :
            joinTable = cls.getTable(joinTable)
            joinObject = Join(validType, Table.table, joinTable)
            Table.table = joinTable
        return joinObject

    @classmethod
    def getType(cls, joinType) -> int :
        """Get a valid join type option from input join type."""
        if isinstance(joinType, int) :
            validType = joinType
            if joinType < 0 or joinType > 4 : validType = 0
        else :
            if joinType == 'INNER JOIN' or joinType == 'INNER' :
                validType = Join.INNER_JOIN
            elif joinType == 'LEFT JOIN' or joinType == 'LEFT' :
                validType = Join.LEFT_JOIN
            elif joinType == 'RIGHT JOIN' or joinType == 'RIGHT' :
                validType = Join.RIGHT_JOIN
            elif joinType == 'OUTER JOIN' or joinType == 'OUTER' :
                validType = Join.OUTER_JOIN
            else :
                validType = Join.NO_JOIN
        return validType

    @classmethod
    def getTable(cls, table) -> str :
        valiTable = ''
        if isinstance(table, str) :
            valiTable += table
        elif isinstance(table, bytes) or isinstance(table, bytearray) :
            valiTable += str(table, 'utf-8')
        return valiTable

    def addColumn(self, baseColumn, joinColumn = None) :
        """Add columns object property to a join table object."""
        table = Table.table
        Table.table = Join.table
        baseColumnObject = Column.create(baseColumn)
        Table.table = table
        if joinColumn is None :
            self.__usingColumns += (baseColumnObject,)
        else :
            joinColumnObject = Column.create(joinColumn)
            self.__baseColumns += (baseColumnObject,)
            self.__joinColumns += (joinColumnObject,)



from ...Structure import Clause
from ...Builder import WhereBuilder, HavingBuilder

class Clauses :
    """Clause manipulation component.
    Used for WHERE and HAVING query.
    """

    def __init__(self) :
        Clause.clauseType = Clause.NONE
        Clause.nestedConjunctive = Clause.CONJUNCTIVE_NONE
        Clause.nestedLevel = 0

    def __clauses(self, column, operator: int, value, conjunctive: int) :
        """Add Clause object to where or having property of builder object"""
        clauseType = Clause.clauseType
        if clauseType == Clause.NONE : clauseType = Clause.WHERE
        clauseObject = Clause.create(clauseType, column, operator, value, conjunctive)
        if isinstance(self.builder, WhereBuilder) or isinstance(self.builder, HavingBuilder) :
            if Clause.clauseType == Clause.HAVING :
                self.builder.addHaving(clauseObject)
            else :
                self.builder.addWhere(clauseObject)
        else :
            raise Exception('Builder object does not support WHERE or HAVING query')
        return self

    def equal(self, column, value) :
        """WHERE or HAVING clause query with "=" operator and AND conjunctive"""
        return self.__clauses(column, Clause.OPERATOR_EQUAL, value, Clause.CONJUNCTIVE_AND)

    def orEqual(self, column, value) :
        """WHERE or HAVING clause query with "=" operator and OR conjunctive"""
        return self.__clauses(column, Clause.OPERATOR_EQUAL, value, Clause.CONJUNCTIVE_OR)

    def notEqual(self, column, value) :
        """WHERE or HAVING clause query with "!=" operator and AND conjunctive"""
        return self.__clauses(column, Clause.OPERATOR_NOT_EQUAL, value, Clause.CONJUNCTIVE_AND)

    def notOrEqual(self, column, value) :
        """WHERE or HAVING clause query with "!=" operator and OR conjunctive"""
        return self.__clauses(column, Clause.OPERATOR_NOT_EQUAL, value, Clause.CONJUNCTIVE_OR)

    def greater(self, column, value) :
        """WHERE or HAVING clause query with ">" operator and AND conjunctive"""
        return self.__clauses(column, Clause.OPERATOR_GREATER, value, Clause.CONJUNCTIVE_AND)

    def orGreater(self, column, value) :
        """WHERE or HAVING clause query with ">" operator and OR conjunctive"""
        return self.__clauses(column, Clause.OPERATOR_GREATER, value, Clause.CONJUNCTIVE_OR)

    def greaterEqual(self, column, value) :
        """WHERE or HAVING clause query with ">=" operator and AND conjunctive"""
        return self.__clauses(column, Clause.OPERATOR_GREATER_EQUAL, value, Clause.CONJUNCTIVE_AND)

    def orGreaterEqual(self, column, value) :
        """WHERE or HAVING clause query with ">=" operator and OR conjunctive"""
        return self.__clauses(column, Clause.OPERATOR_GREATER_EQUAL, value, Clause.CONJUNCTIVE_OR)

    def less(self, column, value) :
        """WHERE or HAVING clause query with "<" operator and AND conjunctive"""
        return self.__clauses(column, Clause.OPERATOR_LESS, value, Clause.CONJUNCTIVE_AND)

    def orLess(self, column, value) :
        """WHERE or HAVING clause query with "<" operator and OR conjunctive"""
        return self.__clauses(column, Clause.OPERATOR_LESS, value, Clause.CONJUNCTIVE_OR)

    def lessEqual(self, column, value) :
        """WHERE or HAVING clause query with "<=" operator and AND conjunctive"""
        return self.__clauses(column, Clause.OPERATOR_LESS_EQUAL, value, Clause.CONJUNCTIVE_AND)

    def orLessEqual(self, column, value) :
        """WHERE or HAVING clause query with "<=" operator and OR conjunctive"""
        return self.__clauses(column, Clause.OPERATOR_LESS_EQUAL, value, Clause.CONJUNCTIVE_OR)

    def between(self, column, value) :
        """WHERE or HAVING clause query with BETWEEN operator and AND conjunctive"""
        return self.__clauses(column, Clause.OPERATOR_BETWEEN, value, Clause.CONJUNCTIVE_AND)

    def orBetween(self, column, value) :
        """WHERE or HAVING clause query with BETWEEN operator and OR conjunctive"""
        return self.__clauses(column, Clause.OPERATOR_BETWEEN, value, Clause.CONJUNCTIVE_OR)

    def notBetween(self, column, value) :
        """WHERE or HAVING clause query with NOT BETWEEN operator and AND conjunctive"""
        return self.__clauses(column, Clause.OPERATOR_NOT_BETWEEN, value, Clause.CONJUNCTIVE_AND)

    def notOrBetween(self, column, value) :
        """WHERE or HAVING clause query with NOT BETWEEN operator and OR conjunctive"""
        return self.__clauses(column, Clause.OPERATOR_NOT_BETWEEN, value, Clause.CONJUNCTIVE_OR)

    def In(self, column, value) :
        """WHERE or HAVING clause query with IN operator and AND conjunctive"""
        return self.__clauses(column, Clause.OPERATOR_IN, value, Clause.CONJUNCTIVE_AND)

    def orIn(self, column, value) :
        """WHERE or HAVING clause query with IN operator and OR conjunctive"""
        return self.__clauses(column, Clause.OPERATOR_IN, value, Clause.CONJUNCTIVE_OR)

    def notIn(self, column, value) :
        """WHERE or HAVING clause query with NOT IN operator and AND conjunctive"""
        return self.__clauses(column, Clause.OPERATOR_NOT_IN, value, Clause.CONJUNCTIVE_AND)

    def notOrIn(self, column, value) :
        """WHERE or HAVING clause query with NOT IN operator and OR conjunctive"""
        return self.__clauses(column, Clause.OPERATOR_NOT_IN, value, Clause.CONJUNCTIVE_OR)

    def isNull(self, column) :
        """WHERE or HAVING clause query with NULL operator and AND conjunctive"""
        return self.__clauses(column, Clause.OPERATOR_NULL, None, Clause.CONJUNCTIVE_AND)

    def orIsNull(self, column) :
        """WHERE or HAVING clause query with NULL operator and OR conjunctive"""
        return self.__clauses(column, Clause.OPERATOR_NULL, None, Clause.CONJUNCTIVE_OR)

    def notIsNull(self, column) :
        """WHERE or HAVING clause query with NOT NULL operator and AND conjunctive"""
        return self.__clauses(column, Clause.OPERATOR_NOT_NULL, None, Clause.CONJUNCTIVE_AND)

    def notOrIsNull(self, column) :
        """WHERE or HAVING clause query with NOT NULL operator and OR conjunctive"""
        return self.__clauses(column, Clause.OPERATOR_NOT_NULL, None, Clause.CONJUNCTIVE_OR)



from ...Structure import Limit
from ...Builder import LimitBuilder

class LimitOffset :
    """LIMIT or OFFSET manipulation component.
    """

    def limit(self, limit, offset = Limit.NOT_SET) :
        """LIMIT or OFFSET query manipulation"""
        limitObject = Limit.create(limit, offset)
        if isinstance(self.builder, LimitBuilder) :
            self.builder.setLimit(limitObject)
        else :
            raise Exception('Builder object does not support LIMIT and OFFSET query')
        return self

    def offset(self, offset) :
        """OFFSET query manipulation"""
        return self.limit(Limit.NOT_SET, offset)



from ..QueryObject import QueryObject
from ..Builder import BaseBuilder
from ..Structure import Table, Column, Value, Clause, Order, Limit, Expression, Join

class BaseTranslator :
    """Translator base class.
    Contain methods for translating builder object to query object.
    Methods and class properties can be overload by a derived class for creating spesific database translator.
    """

    def __init__(self) :
        self.quote_struct = "`"
        self.quote_string = "'"
        self.equal = "="
        self.open_bracket = "("
        self.close_bracket = ")"
        self.dot = "."
        self.comma = ","
        self.end_query = ""

    def firstKeyword(self, query: QueryObject, builderType: int) :
        """Create first query keyword from builder type."""
        if builderType == BaseBuilder.SELECT :
            query.add('SELECT ')
        elif builderType == BaseBuilder.INSERT :
            query.add('INSERT ')
        elif builderType == BaseBuilder.UPDATE :
            query.add('UPDATE ')
        elif builderType == BaseBuilder.DELETE :
            query.add('DELETE')
        elif builderType == BaseBuilder.SELECT_DISTINCT :
            query.add('SELECT DISTINCT ')

    def fromTable(self, query: QueryObject, table: Table) :
        """Create "FROM `table` AS `alias`" expression"""
        name = table.name()
        alias = table.alias()
        query.add(' FROM ' + self.quote_struct)
        query.add(name)
        query.add(self.quote_struct)
        if alias :
            query.add(' AS ' + self.quote_string + alias + self.quote_string)

    def intoTable(self, query: QueryObject, table: Table) :
        """Create "INTO `table`" expression"""
        name = table.name()
        query.add('INTO ' + self.quote_struct)
        query.add(name)
        query.add(self.quote_struct)

    def tableSet(self, query: QueryObject, table: Table) :
        """Create "`table` SET" expression"""
        name = table.name()
        query.add(self.quote_struct)
        query.add(name)
        query.add(self.quote_struct)

    def columnsSelect(self, query: QueryObject, columns: tuple, count: int, multiTableFlag: bool = False) :
        """Generate list of "FUNCTION(`table`.`column`) AS (`alias`)" for SELECT query"""
        if count == 0 :
            query.add('*')
            return
        for column in columns :
            if isinstance(column, Column) :
                table = column.table()
                name = column.name()
                alias = column.alias()
                function = column.function()
                if function :
                    query.add(function + self.open_bracket)
                query.add(self.quote_struct)
                if table and multiTableFlag :
                    query.add(table)
                    query.add(self.quote_struct + self.dot + self.quote_struct)
                query.add(name)
                query.add(self.quote_struct)
                if function :
                    query.add(self.close_bracket)
                if alias :
                    query.add(' AS ' + self.quote_string + alias + self.quote_string)
            elif isinstance(column, Expression) :
                self.expression(query, column)
            count -= 1
            if count > 0 : query.add(self.comma)

    def columnsInsert(self, query: QueryObject, values: tuple, count: int) :
        """Generate list of "`column`" expression for INSERT query"""
        if count == 0 :
            query.add(' ' + self.open_bracket)
            query.add(self.close_bracket)
            return
        value = values[0]
        if isinstance(value, Value) :
            columns = value.columns()
            count = len(columns)
            query.add(' ' + self.open_bracket)
            for column in columns : 
                query.add(self.quote_struct)
                query.add(column)
                query.add(self.quote_struct)
                count -= 1
                if count > 0 : query.add(self.comma)
            query.add(self.close_bracket)

    def column(self, query: QueryObject, column: Column, multiTableFlag: bool = False) :
        """Create "FUNCTION(`table`.`column`)" or "FUNCTION(`alias`)" expression"""
        table = column.table()
        name = column.name()
        alias = column.alias()
        function = column.function()
        if function :
            query.add(function + self.open_bracket)
        if alias :
            query.add(self.quote_struct)
            query.add(column.alias())
            query.add(self.quote_struct)
        else :
            query.add(self.quote_struct)
            if table and multiTableFlag :
                query.add(table)
                query.add(self.quote_struct + self.dot + self.quote_struct)
            query.add(name)
            query.add(self.quote_struct)
        if function :
            query.add(self.close_bracket)

    def valuesInsert(self, query: QueryObject, values: tuple, count: int) :
        """Generate list of "FUNCTION(`table`.`column`)" or "FUNCTION(`alias`)" expression for INSERT query"""
        query.add(' VALUES ')
        if count == 0 :
            query.add(self.open_bracket)
            query.add(self.close_bracket)
            return
        for value in values :
            if isinstance(value, Value) :
                vals = value.values()
                countVals = len(vals)
                query.add(self.open_bracket)
                for val in vals :
                    query.add(val, True)
                    countVals -= 1
                    if countVals > 0 : query.add(self.comma)
                query.add(self.close_bracket)
            count -= 1
            if count > 0 : query.add(self.comma)

    def valuesUpdate(self, query: QueryObject, values: tuple, count: int, multiTableFlag: bool = False) :
        """Generate list of "`table`.`column`=values" expression for UPDATE query"""
        query.add(' SET ')
        for value in values :
            if isinstance(value, Value) :
                table = value.table()
                columns = value.columns()
                vals = value.values()
                countVals = len(vals)
                for i, val in enumerate(vals) :
                    query.add(self.quote_struct)
                    if multiTableFlag :
                        query.add(table)
                        query.add(self.quote_struct + self.dot + self.quote_struct)
                    query.add(columns[i])
                    query.add(self.quote_struct + self.equal)
                    query.add(val, True)
                    countVals -= 1
                    if countVals > 0 : query.add(self.comma)
            count -= 1
            if count > 0 : query.add(self.comma)

    def expression(self, query: QueryObject, expression: Expression) :
        """Create user defined expression with param binding and alias."""
        params = expression.params()
        exps = expression.expression()
        alias = expression.alias()
        for i in range(len(exps)) :
            query.add(exps[i])
            if params[i] is not None : query.add(params[i], True)
        if alias :
            query.add(' AS ' + self.quote_struct)
            query.add(alias)
            query.add(self.quote_struct)

    def join(self, query: QueryObject, joins: tuple) :
        """Generate JOIN expression with join table and columns"""
        for join in joins :
            if isinstance(join, Join) :
                joinType = self.joinType(join.joinType())
                query.add(joinType)
                self.joinTable(query, join.joinTable(), join.joinAlias())
                self.joinColumns(query, join.baseColumns(), join.joinColumns(), join.usingColumns())

    def joinType(self, joinType: int) -> str :
        """Translate join type for join query"""
        if joinType == Join.INNER_JOIN :
            return ' INNER JOIN '
        elif joinType == Join.LEFT_JOIN :
            return ' LEFT JOIN '
        elif joinType == Join.RIGHT_JOIN :
            return ' RIGHT JOIN '
        elif joinType == Join.OUTER_JOIN :
            return ' OUTER JOIN '
        else :
            return ''

    def joinTable(self, query: QueryObject, joinTable: str, joinAlias: str) :
        """Create "`table` AS `alias`" expression for JOIN query"""
        query.add(self.quote_struct)
        query.add(joinTable)
        if joinAlias :
            query.add(self.quote_struct + ' AS ' + self.quote_struct + joinAlias)
        query.add(self.quote_struct)

    def joinColumns(self, query: QueryObject, baseColumns: tuple, joinColumns: tuple, usingColumns: tuple) :
        """Create "USING(`column`)" or "ON `basetable`.`column`=`jointable`.`column`" expression for JOIN query"""
        count = len(usingColumns)
        if count :
            query.add(' USING ' + self.open_bracket)
            for column in usingColumns :
                self.column(query, column)
                count -= 1
                if count > 0 : query.add(self.comma)
            query.add(self.close_bracket)
        else :
            for i in range(len(baseColumns)) :
                if i == 0 : query.add(' ON ') 
                else : query.add(' AND ')
                self.column(query, baseColumns[i], True)
                query.add(self.equal)
                self.column(query, joinColumns[i], True)

    def operator(self, operator: int) -> str :
        """Translate operator for WHERE and HAVING query and comparison expression"""
        if operator == Clause.OPERATOR_EQUAL :
            return '='
        elif operator == Clause.OPERATOR_NOT_EQUAL :
            return '!='
        elif operator == Clause.OPERATOR_GREATER :
            return '>'
        elif operator == Clause.OPERATOR_GREATER_EQUAL :
            return '>='
        elif operator == Clause.OPERATOR_LESS :
            return '<'
        elif operator == Clause.OPERATOR_LESS_EQUAL :
            return '<='
        elif operator == Clause.OPERATOR_BETWEEN :
            return ' BETWEEN '
        elif operator == Clause.OPERATOR_NOT_BETWEEN :
            return ' NOT BETWEEN '
        elif operator == Clause.OPERATOR_LIKE :
            return ' LIKE '
        elif operator == Clause.OPERATOR_NOT_LIKE :
            return ' NOT LIKE '
        elif operator == Clause.OPERATOR_IN :
            return ' IN '
        elif operator == Clause.OPERATOR_NOT_IN :
            return ' NOT IN '
        elif operator == Clause.OPERATOR_NULL :
            return ' IS NULL '
        elif operator == Clause.OPERATOR_NOT_NULL :
            return ' IS NOT NULL '
        else :
            return ''

    def conjunctive(self, conjunctive: int) -> str :
        """Translate conjunctive for WHERE and HAVING clause query"""
        if conjunctive == Clause.CONJUNCTIVE_AND :
            return ' AND '
        elif conjunctive == Clause.CONJUNCTIVE_OR :
            return ' OR '
        elif conjunctive == Clause.CONJUNCTIVE_NOT_AND :
            return ' NOT AND '
        elif conjunctive == Clause.CONJUNCTIVE_NOT_OR :
            return ' NOT OR '
        else :
            return ''

    def brackets(self, level: int) -> str :
        """Get open or close bracket based on input level"""
        string = ''
        if level < 0 :
            for i in range(level, 0) :
                string += self.open_bracket
        elif level > 0 :
            for i in range(0, level) :
                string += self.close_bracket
        return string

    def where(self, query: QueryObject, whereClauses: tuple, count: int, multiTableFlag: bool = False) :
        """Create "WHERE" expression and generate list of clause expression"""
        if count :
            query.add(' WHERE ')
            for where in whereClauses :
                if isinstance(where, Clause) :
                    conjunctive = self.conjunctive(where.conjunctive())
                    nestedLevel = where.level()
                    query.add(conjunctive)
                    if nestedLevel < 0 : query.add(self.brackets(nestedLevel))
                    self.clause(query, where, multiTableFlag)
                    if nestedLevel > 0 : query.add(self.brackets(nestedLevel))

    def having(self, query: QueryObject, havingClauses: tuple, count: int, multiTableFlag: bool = False) :
        """Create "HAVING" expression and generate list of clause expression"""
        if count :
            query.add(' HAVING ')
            for having in havingClauses :
                if isinstance(having, Clause) :
                    conjunctive = self.conjunctive(having.conjunctive())
                    nestedLevel = having.level()
                    query.add(conjunctive)
                    if nestedLevel < 0 : query.add(self.brackets(nestedLevel))
                    self.clause(query, having, multiTableFlag)
                    if nestedLevel > 0 : query.add(self.brackets(nestedLevel))

    def clause(self, query: QueryObject, clause: Clause, multiTableFlag: bool = False) :
        """Create clause expression which contain column, operator, and value"""
        column = clause.column()
        operator = clause.operator()
        value = clause.value()
        if isinstance(column, Column) :
            self.column(query, column, multiTableFlag)
        elif isinstance(column, Expression) :
            self.expression(query, column)
        if operator == Clause.OPERATOR_BETWEEN or operator == Clause.OPERATOR_NOT_BETWEEN :
            self.clauseBetween(query, value)
        elif operator == Clause.OPERATOR_IN or operator == Clause.OPERATOR_NOT_IN :
            self.clauseIn(query, value)
        else :
            self.clauseComparison(query, value, operator)

    def clauseComparison(self, query: QueryObject, value, operator: int) :
        """Create comparison expression in clause expression"""
        query.add(self.operator(operator))
        if operator != Clause.OPERATOR_NULL and operator != Clause.OPERATOR_NOT_NULL :
            query.add(value, True)

    def clauseBetween(self, query: QueryObject, value) :
        """Create between expression in clause expression"""
        query.add(self.operator(Clause.OPERATOR_BETWEEN))
        query.add(value[0], True)
        query.add(' AND ')
        query.add(value[1], True)

    def clauseIn(self, query: QueryObject, value) :
        """Create in expression in clause expression"""
        query.add(self.operator(Clause.OPERATOR_IN))
        query.add(self.open_bracket)
        count = len(value)
        for i in range(count) :
            query.add(value[i], True)
            if i < count - 1 : query.add(self.comma)
        query.add(self.close_bracket)

    def groupBy(self, query: QueryObject, groups: tuple, count: int, multiTableFlag: bool = False) :
        """Create "GROUP BY " expression and generate list of column expression"""
        if count :
            query.add(' GROUP BY ')
            for group in groups :
                if isinstance(group, Column) :
                    self.column(query, group, multiTableFlag)
                    count -= 1
                    if count > 0 : query.add(self.comma)

    def orderBy(self, query: QueryObject, orders: tuple, count: int, multiTableFlag: bool = False) :
        """Create "ORDER BY `column` ASC|DESC" expression"""
        if count :
            query.add(' ORDER BY ')
            for order in orders :
                if isinstance(order, Order) :
                    self.column(query, order.column(), multiTableFlag)
                    if (order.orderType() == Order.ORDER_ASC) :
                        query.add(' ASC')
                    elif (order.orderType() == Order.ORDER_DESC) :
                        query.add(' DESC')
                    count -= 1
                    if count > 0 : query.add(self.comma)

    def limitOffset(self, query: QueryObject, limitOffset: Limit, hasLimit: bool) :
        """Create "OFFSET offsetValue" or "LIMIT limitValue, offsetValue" expression"""
        if hasLimit :
            limit = limitOffset.limit()
            offset = limitOffset.offset()
            if limit == Limit.NOT_SET :
                query.add(' OFFSET ')
                query.add(offset, True)
            else :
                query.add(' LIMIT ')
                query.add(limit, True)
                if offset != Limit.NOT_SET :
                    query.add(self.comma)
                    query.add(offset, True)



from .BaseBuilder import BaseBuilder
from .SelectBuilder import SelectBuilder
from .InsertBuilder import InsertBuilder
from .UpdateBuilder import UpdateBuilder
from .DeleteBuilder import DeleteBuilder
from .Component import WhereBuilder
from .Component import HavingBuilder
from .Component import GroupByBuilder
from .Component import OrderByBuilder
from .Component import LimitBuilder
from .Component import JoinBuilder


from .BaseBuilder import BaseBuilder
from .Component import WhereBuilder, LimitBuilder, JoinBuilder

class UpdateBuilder(BaseBuilder, WhereBuilder, LimitBuilder, JoinBuilder) :
    """Template for building UPDATE query
    Components:
    - Where builder
    - Limit builder
    - Join builder
    """

    def __init__(self) :
        BaseBuilder.__init__(self)
        WhereBuilder.__init__(self)
        LimitBuilder.__init__(self)
        JoinBuilder.__init__(self)
